project = "pgqueuer"
copyright = "2024, JeeyBee"
author = "JeeyBee"
extensions = ["myst_parser"]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_static_path = ["_static"]
pygments_style = "sphinx"
html_theme = "sphinx_rtd_theme"
